#include <cstdio>
#include <cstring>
#include <cmath>

using namespace std;

#include "snd.hh"
#include "instrument.hh"
#include "sndplayer.hh"
#include "analysis.hh"
#include "s3m.hh"
#include "adlmidi.h"
#include "midi-gm.h"
#include "midi.hh"

static void DumpSND(const song &song)
{
    for(unsigned a=0; a<song.rownum; ++a)
    {
        for(unsigned c=0; c<song.channelcount; ++c)
        {
            const SNDChannel &tmp = song.ChanPtr(a, c);
            char Buf[64];
            int colormode=0;
            
            sprintf(Buf, "v%Xs%dw%ua%u", tmp.vol, tmp.sample, tmp.wavelen, tmp.age);
            if(tmp.ins)
                sprintf(strchr(Buf,0), "(%u)", tmp.ins);
            
            sprintf(Buf+32, "%18s", Buf);
            if(tmp.vol)
            {
                if(IsBeat(a, c))
                {
                    fputs("\33[1;34m", stderr);
                    colormode=2;
                }
                else
                    colormode=1;
            }
            else
            {
                colormode=0;
                fputs("\33[1;30m", stderr);
            }
            for(char *s=Buf+32; *s; ++s)
            {
                /*
                if(colormode==1)
                  if(*s=='v'||*s=='s'||*s=='w'||*s=='a')fputs("\33[1;37m", stderr);
                */
                fputc(*s, stderr    );
                /*
                if(colormode==1)
                  if(*s=='v'||*s=='s'||*s=='w'||*s=='a')fputs("\33[m", stderr);
                */
            }
            fputs("\33[m", stderr);
        }
        fprintf(stderr, "\n");
    }
}


static unsigned NesVoltoS3Mvol(unsigned v)
{
    static unsigned char voltable[16] =
    {0,32,38,40,
     42,46,48,50,
     52,54,56,58,
     60,62,63,64};
    return voltable[v];
}

static void Patternize(const song &song)
{
    S3M musicfile;
    S3M::eventtable events;
    
    events.setgeometry(0, song.channelcount);
    
    /* Handle each channel separately */
    for(unsigned chan=0; chan<song.channelcount; ++chan)
    {
        vector<S3M::event> chanevents;
        
        switch(song.ChanPtr(0, chan).type)
        {
            case SNDChannel::pcm:
                musicfile.setupchannel(chan, S3M::leftchan(chan));
                break;
            default:
                musicfile.setupchannel(chan, S3M::amechan(chan));
                break;
        }
        
        unsigned lastrow=0;
        bool wason = false;
        for(double pos=0; ; pos += TickSpacer)
        {
            unsigned row = (unsigned)pos;
            if(row >= song.rownum)break;
        
            const SNDChannel &tmp = song.ChanPtr(row, chan);
            
            // vol=tmp.vol     (volume(0..15))
            // smp=tmp.sample  (sqr:duty; pcm:pcm sample no)
            // wav=tmp.wavelen (wave length)
            // age=tmp.age     (age since last touched)
            // ins=tmp.ins     (index to repository)
            // beat=beat
            const SNDChannel *beatrow = NULL;
            while(lastrow <= row)
                if(IsBeat(lastrow++, chan))
                    beatrow = &song.ChanPtr(lastrow-1, chan);
            
            S3M::event ev;
            
            if(beatrow)
            {
                double div = 7000.0;
                if(tmp.type == SNDChannel::rnd)
                {
                    div /= 32;
                }
                else if(tmp.type == SNDChannel::pcm)
                {
                    div /= 4;
                }
                else if(tmp.type == SNDChannel::tri)
                {
                    div /= 8;
                }
                unsigned basewave = tmp.wavelen;
                unsigned nuotti = (unsigned)(12 * log(div / basewave) / log(2.0));
                unsigned oct = nuotti/12;
                ev.note = (nuotti%12) + oct*16;
                wason = true;
            }
            else if(wason && !tmp.vol)
            {
                ev.note = 254;
                ev.vol = 0;
                wason = false;
            }
            
            if(tmp.type == SNDChannel::pcm && beatrow && tmp.sample)
                ev.ins = tmp.sample + repository.size();
            else if(tmp.type == SNDChannel::pcm && beatrow)
                ev.ins = beatrow->sample + repository.size();
            else if(tmp.ins)
                ev.ins = tmp.ins;
            else if(beatrow && beatrow->ins)
                ev.ins = beatrow->ins;

            if(beatrow)
            {
                int volavg = 0, volcount=0;
                for(unsigned tmprow = row; tmprow < row+8 && tmprow < song.rownum; ++tmprow)
                {
                    if(tmprow > row && IsBeat(tmprow, chan))break;
                    const SNDChannel &tmp = song.ChanPtr(tmprow, chan);
                    if(!tmp.vol)break;
                    volavg += tmp.vol; ++volcount;
                }
                if(volcount) volavg /= volcount;
                volavg = NesVoltoS3Mvol(volavg);
                
                if(tmp.type != SNDChannel::pcm && tmp.ins)
                {
                    if(!repository[tmp.ins-1].defaultvol)
                        repository[tmp.ins-1].defaultvol = volavg;
                    else if(repository[tmp.ins-1].defaultvol != volavg)
                        ev.vol = volavg;
                }
                else
                    ev.vol = volavg;
            }
            else
            {
                //ev.vol = NesVoltoS3Mvol(tmp.vol);
            }
            
            chanevents.push_back(ev);
        }
        events.loadchannel(chan, chanevents);
    }

    fprintf(stderr, "----------\n");
    for(unsigned a=0; a<events.rowcount; ++a)
    {
        fprintf(stderr, "%u ", a % 4);
        for(unsigned c=0; c<events.channelcount; ++c)
        {
            const S3M::event& ev = events.events[c + events.channelcount * a];
            
            switch(ev.note)
            {
                case 255: fprintf(stderr, "..."); break;
                case 254: fprintf(stderr, "^^^"); break;
                default:
                    char Buf[64];
                    static const char nuotit[12*3+1] = "C-\0C#\0D-\0D#\0E-\0F-\0F#\0G-\0G#\0A-\0A#\0B-\0";
                    sprintf(Buf, "%s%d", nuotit+3*(ev.note%16), ev.note/16);
                    fprintf(stderr, "%s", Buf);
            }
            if(ev.ins) fprintf(stderr, " %02u", ev.ins);
            else fprintf(stderr, " ..");
            if(ev.vol != 255) fprintf(stderr, " %02u", ev.vol);
            else fprintf(stderr, " ..");
            
            fprintf(stderr, "  ");
        }
        fprintf(stderr, "\n");
    }
    unsigned patcount = (events.rowcount + 64 - 1) / 64;
    for(unsigned a=0; a<patcount; ++a)
    {
        S3M::eventtable tmp(events, a*64, 64);
        musicfile.getpattern(a) = musicfile.compresspattern(tmp, 64);
        musicfile.setorder(a, a);
    }
    musicfile.setorder(patcount, 255);
    musicfile.setorder(patcount+1, 255);
    
    for(unsigned a=1; a<256; ++a)
        if(song.dpcmcache[a].size())
        {
            S3M::instrument &ins = musicfile.getinstrument(a + repository.size() - 1);
            
            ins.MakeSMP();
            ins.audio = song.dpcmcache[a];
            for(unsigned c=0; c<ins.audio.size(); ++c)
            {
                ins.audio[c] = 0x80 + (ins.audio[c]-0x80)*4;
            }
            ins.SetLoop(0, ins.audio.size());

            ins.name = "GM1/0:";
        }

    for(unsigned a=0; a<repository.size(); ++a)
    {
        sample& smp = repository[a];
        
        S3M::instrument& ins = musicfile.getinstrument(a);
        ins.MakeAME();
        for(unsigned d=0; d<11; ++d)
            ins.data[d] = OPL_GMdata(smp.midiins, d);
        ins.data[12] = smp.defaultvol;
        
        ins.name = MidiInsName[smp.midiins];
    }
    
    musicfile.Save("test.s3m");
}

static unsigned NesVoltoMIDIvol(unsigned v)
{
    return (34-pow(15-v, 1.3)) * 128 / 35;
}

static void Midize(const song &song)
{
    MIDIfile midifile;
    
    /* Handle each channel separately */
    for(unsigned chan=0; chan<song.channelcount; ++chan)
    {
        midifile.InitTrack(chan);
        
        unsigned lastrow=0;
        
        bool wason = false;
        
        unsigned actchan=255;
        unsigned actnote=255;
        unsigned actpatch=255;
        unsigned actvol=255;
        int transpone = 0;
        int resultvol = 100;
        
        for(unsigned row=0; row<song.rownum; ++row)
        {
        	/*if(!(row&1))midifile.AddDelay(chan, 35);*/

            const SNDChannel &tmp = song.ChanPtr(row, chan);
            
            if(tmp.type == SNDChannel::pcm)
            {
                // FIXME: For now, we ignore PCM in MIDI
                continue;
            }
            
            // vol=tmp.vol     (volume(0..15))
            // smp=tmp.sample  (sqr:duty; pcm:pcm sample no)
            // wav=tmp.wavelen (wave length)
            // age=tmp.age     (age since last touched)
            // ins=tmp.ins     (index to repository)
            // beat=beat
            
            bool beat = IsBeat(row, chan);
            
            if(wason && (beat || !tmp.vol))
            {
                // Deactivate instrument
                midifile.KeyOff(chan, actchan, actnote, 60);
                wason = false;
            }

            if(beat)
            {
                if(tmp.ins)
                {
                    unsigned midiins = repository[tmp.ins-1].midiins;
                    transpone        = repository[tmp.ins-1].transpone;
                    resultvol        = repository[tmp.ins-1].resultvolume;
                    
                    // Set patch (if changed)
                    if(midiins < 128)
                    {
                        if(actpatch != midiins)
                        {
                            if(actchan != chan)actvol = 255;
                            actchan  = chan;
                            actpatch = midiins;
                            midifile.Patch(chan, actchan, actpatch);
                        }
                    }
                    else
                    {
                        if(actchan != 9)actvol = 255;
                        actchan  = 9;
                        actpatch = midiins;
                        actnote  = actpatch-128+35;
                    }
                }
                
                int volavg = 0, volcount=0;
                for(unsigned tmprow = row; tmprow < row+8 && tmprow < song.rownum; ++tmprow)
                {
                    if(tmprow > row && IsBeat(tmprow, chan))break;
                    const SNDChannel &tmp = song.ChanPtr(tmprow, chan);
                    int vol = tmp.vol;
                    if(!vol)break;
                    if(tmp.type == SNDChannel::tri)
                    {
                    	vol = vol * 6 / 10;
                    }                    
                    volavg += vol; ++volcount;
                }
                volavg /= volcount;
                volavg = volavg * resultvol / 100;
                volavg = NesVoltoMIDIvol(volavg);
                
                // Set channel volume (if changed)
                
                if(actvol != volavg)
                {
                    midifile.SetVolume(chan, actchan, volavg);
                    actvol = volavg;
                }
            }
            else if(tmp.vol)
            {
            	/*
                int voltmp = NesVoltoMIDIvol(tmp.vol);
                voltmp = voltmp * resultvol / 100;

                // Set channel volume (if changed)
                
                if(actvol != voltmp)
                {
                    midifile.SetVolume(chan, actchan, voltmp);
                    actvol = voltmp;
                }
                */
            }
            
            if(beat)
            {
                double div = 20000.0;
                if(tmp.type == SNDChannel::rnd)
                {
                    div /= 32;
                }
                else if(tmp.type == SNDChannel::pcm)
                {
                    div /= 4;
                }
                else if(tmp.type == SNDChannel::tri)
                {
                    div /= 2;
                }
                unsigned basewave = tmp.wavelen;
                unsigned nuotti =
                  (unsigned)(12 * log(div / basewave) / log(2.0)) - 18 + transpone;
                
                // Active instrument
                if(actchan == 9)
                {
                    midifile.KeyOn(chan, actchan, actnote, 100);
                }
                else
                {
                    actnote = nuotti;
                    midifile.KeyOn(chan, actchan, actnote, 100);
                }
                
                wason = true;
            }
            else if(tmp.vol)
            {
                // TODO: Check pitch bending
            }
        	midifile.AddDelay(chan, 18);
        }
    }

    midifile.Done();
    
    midifile.Save("test.mid");
}

int main(int argc, const char *const *argv)
{
    if(argc != 2 || !strcmp(argv[1], "--help") || !strcmp(argv[1], "/?"))
    {
        puts("Simple SNDv3 player - Copyright (C) 1992,2006 Bisqwit (http://iki.fi/bisqwit/)\n"
             "Usage:\n"
             "   snd <filename.snd>\n"
             "\n"
             " Note: This movie will write\n"
             "   lots of debugging output to stderr\n"
             "   and raw PCM data into stdout!\n"
             "\n"
             " So be sure to redirect them somewhere.\n"
            );
        return 1;
    }
    
    song song = ReadSND(argv[1]);
    
    /* TODO: Handle the snd (dump it somewhere or something) */
    
    AnalyzeSND(song);

    for(unsigned a=0; a<repository.size(); ++a)
    {
        fprintf(stderr, "Instrument %u:\n", a+1);
        GuessMIDIfor(repository[a]);
        DumpSample(repository[a]);
    }
    
    DumpSND(song);
    
    //TickSpacer = 7 / 2.0;
    if(TickSpacer == 1.0)
        TickSpacer = 2.0;
    
    Patternize(song);
    Midize(song);
    PlaySND(song);
    
    return 0;
}
